import collections
import copy
import datetime
import locale
import time
import weakref

from marnadi.utils import cached_property, CachedDescriptor


class CookieJar(collections.MutableMapping):
    """Cookies - dict-like object allowing to get/set HTTP cookies"""

    if hasattr(collections.MutableMapping, '__slots__'):
        __slots__ = ('_response', 'domain', 'path', 'expires', 'secure',
                     'http_only', '__weakref__')

    def __init__(self, response, domain=None, path=None, expires=None,
                 secure=False, http_only=True, ):
        self._response = weakref.ref(response)
        self.domain = domain
        self.path = path
        self.expires = expires
        self.secure = secure
        self.http_only = http_only

    __hash__ = object.__hash__

    __eq__ = object.__eq__

    __ne__ = object.__ne__

    def __setitem__(self, cookie, value):
        self.set(cookie, value)
        if value is None:
            self.request_cookies.pop(cookie, None)
        else:
            self.request_cookies[cookie] = value

    def __delitem__(self, cookie):
        self.remove(cookie)
        del self.request_cookies[cookie]

    def __getitem__(self, cookie):
        return self.request_cookies[cookie]

    def __iter__(self):
        return iter(self.request_cookies)

    def __len__(self):
        return len(self.request_cookies)

    @property
    def response(self):
        response = self._response()
        if response is not None:
            return response
        raise ValueError("CookieJar used outside of response scope")

    @cached_property
    def request_cookies(self):
        try:
            return dict(
                cookie.strip().split('=', 1)
                for cookie in self.response.request.headers['Cookie'].split(';')
            )
        except (KeyError, ValueError):
            return {}

    def clear(self, *cookies):
        if cookies:
            for cookie in cookies:
                self.pop(cookie, None)
        else:
            super(CookieJar, self).clear()

    def remove(self, cookie, domain=None, path=None, secure=None,
               http_only=None):
        self.set(cookie, '', expires=datetime.datetime(1980, 1, 1),
                 domain=domain, path=path, secure=secure, http_only=http_only)

    def set(self, cookie, value, expires=None, domain=None, path=None,
            secure=None, http_only=None):
        if value is None:
            return self.remove(cookie, domain=domain, path=path,
                               secure=secure, http_only=http_only)
        domain = self.domain if domain is None else domain
        path = self.path if path is None else path
        expires = self.expires if expires is None else expires
        secure = self.secure if secure is None else secure
        http_only = self.http_only if http_only is None else http_only

        cookie_params = ['%s=%s' % (cookie, value)]
        domain is not None and cookie_params.append("Domain=%s" % domain)
        path is not None and cookie_params.append("Path=%s" % path)
        if expires is not None:
            try:
                # try to use `expires` as timedelta
                expires += datetime.datetime.now()
            except TypeError:
                pass
            if isinstance(expires, datetime.datetime):
                struct_time = (
                    time.gmtime(time.mktime(expires.timetuple()))
                    if expires.tzinfo is None else
                    time.localtime(time.mktime(expires.utctimetuple()))
                )
                current_locale = locale.getlocale(locale.LC_TIME)
                locale.setlocale(locale.LC_TIME, 'C')
                expires = time.strftime("%a, %d %b %Y %H:%M:%S GMT",
                                        struct_time)
                locale.setlocale(locale.LC_TIME, current_locale)
            cookie_params.append("Expires=%s" % expires)
        secure and cookie_params.append("Secure")
        http_only and cookie_params.append("HttpOnly")
        self.response.headers.append('Set-Cookie', '; '.join(cookie_params))


class Cookies(CachedDescriptor):

    __slots__ = 'domain', 'path', 'expires', 'secure', 'http_only'

    def __init__(self, domain=None, path=None, expires=None, secure=False,
                 http_only=True):
        super(Cookies, self).__init__()
        self.domain = domain
        self.path = path
        self.expires = expires
        self.secure = secure
        self.http_only = http_only

    def get_value(self, response):
        return CookieJar(
            response=response,
            domain=self.domain,
            path=self.path,
            expires=copy.copy(self.expires),
            secure=self.secure,
            http_only=self.http_only,
        )



import collections

from marnadi import Header, descriptors
from marnadi.utils import to_bytes


class HttpError(Exception, collections.Iterable, collections.Sized):

    __slots__ = 'status', 'data', '__weakref__'

    headers = descriptors.Headers(
        ('Content-Type', Header('text/plain', charset='utf-8')),
    )

    def __init__(
        self,
        status='500 Internal Server Error',
        data=None,
        headers=None,
        error=None,
    ):
        self.status = status
        self.data = data or status
        if headers:
            self.headers.extend(*headers)

    def __len__(self):
        return 1

    def __iter__(self):
        yield to_bytes(self.data)



import itertools
import collections

from marnadi.utils import Lazy, CachedDescriptor

from . import decoders


class Data(CachedDescriptor, collections.Mapping):

    __slots__ = '_content_decoders',

    def __init__(self, *content_decoders, **kw_content_decoders):
        super(Data, self).__init__()
        self._content_decoders = {
            content_type: Lazy(content_decoder)
            for content_type, content_decoder in itertools.chain(
                content_decoders,
                kw_content_decoders.items(),
            )
        }

    def __getitem__(self, content_type):
        return self._content_decoders[content_type]

    def __iter__(self):
        return iter(self._content_decoders)

    def __len__(self):
        return len(self._content_decoders)

    def get_value(self, request):
        decoder = self.get(request.content_type, decoders.Decoder)
        return decoder(request)



import importlib
json = importlib.import_module('json')  # import built-in module 'json'

from marnadi.descriptors.data.decoders import Decoder as BaseDecoder


class Decoder(BaseDecoder):

    __slots__ = ('stream', 'encoding')

    def __init__(self):
        self.encoding = None
        self.stream = None

    def __call__(self, request):
        self.encoding = request.content_type.params.get('charset', 'utf-8')
        self.stream = request.input
        return json.load(self)

    def read(self, *args, **kwargs):
        assert self.encoding and self.stream, \
            "Either stream or encoding is not provided"
        return self.stream.read(*args, **kwargs).decode(encoding=self.encoding)



import logging
import itertools

from marnadi import descriptors, Header
from marnadi.errors import HttpError
from marnadi.utils import metaclass, to_bytes, cached_property, coroutine

try:
    str = unicode
except NameError:
    pass


class Handler(type):

    __func__ = None

    logger = logging.getLogger('marnadi')

    def __call__(cls, *args, **kwargs):
        func = cls.__func__
        if func is not None:
            return func(*args, **kwargs)
        return super(Handler, cls).__call__(*args, **kwargs)

    def provider(cls, func):
        assert callable(func)
        attributes = dict(
            __func__=staticmethod(func),
            __module__=func.__module__,
            __doc__=func.__doc__,
            __slots__=(),
        )
        return type(cls)(func.__name__, (cls, ), attributes)

    @coroutine
    def prepare(cls, **kwargs):
        """Initializes response generator with given params.

        Note:
            Error responses can be customized by overriding this method.
            For example your version may catch `HttpError` from original
            implementation and re-raise it with necessary content data
            (which may be a HTML containing formatted traceback).
        """
        app, request = yield
        try:
            response = type.__call__(cls, app, request)  # response instance
            response.start(**kwargs)
            yield response
        except HttpError:
            raise
        except Exception as error:
            cls.logger.exception(error)
            raise HttpError(error=error)


@metaclass(Handler)
class Response(object):

    __slots__ = 'app', 'request', '__weakref__'

    supported_http_methods = {
        'OPTIONS', 'GET', 'HEAD', 'POST', 'PUT', 'PATCH', 'DELETE',
    }

    status = '200 OK'

    headers = descriptors.Headers(
        ('Content-Type', Header('text/plain', charset='utf-8')),
    )

    cookies = descriptors.Cookies()

    def __init__(self, app, request):
        self.app = app
        self.request = request

    def __call__(self, **kwargs):
        if self.request.method not in self.supported_http_methods:
            raise HttpError(
                '501 Not Implemented',
                headers=(('Allow', ', '.join(self.allowed_http_methods)), )
            )
        callback = getattr(
            self,
            self.request.method.lower()
        ) or self.__class__.__func__
        if callback is None:
            raise HttpError(
                '405 Method Not Allowed',
                headers=(('Allow', ', '.join(self.allowed_http_methods)), )
            )
        return callback(**kwargs)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterator)

    def next(self):
        return self.__next__()

    def start(self, **kwargs):
        result = self.__call__(**kwargs)
        self.iterator = itertools.chain(
            (self.iterator.send(result), ),
            self.iterator
        )

    @cached_property
    @coroutine
    def iterator(self):
        result = yield
        if result is None or isinstance(result, (str, bytes)):
            chunk = to_bytes(result)
            self.headers.setdefault('Content-Length', len(chunk))
            yield chunk
        else:
            chunks = iter(result)
            first_chunk = to_bytes(next(chunks, b''))
            try:
                result_length = len(result)
            except TypeError:  # result doesn't support len()
                pass
            else:
                if result_length <= 1:
                    self.headers.setdefault(
                        'Content-Length',
                        len(first_chunk),
                    )
            for chunk in itertools.chain((first_chunk, ), chunks):
                yield to_bytes(chunk, error_callback=self.logger.exception)

    @property
    def allowed_http_methods(self):
        func = self.__class__.__func__
        for method in self.supported_http_methods:
            if func or getattr(self, method.lower()):
                yield method

    def options(self, **kwargs):
        self.headers['Allow'] = ', '.join(self.allowed_http_methods)

    get = None

    head = None

    post = None

    put = None

    patch = None

    delete = None



import re
import collections

from marnadi.utils import Lazy


class Route(object):

    __slots__ = 'path', 'handler', 'params', 'pattern', 'name', 'callbacks'

    placeholder_re = re.compile(r'\{([a-zA-Z_][a-zA-Z0-9_]*)\}')

    def __init__(self, path, handler, name=None, params=None, callbacks=None,
                 patterns=None):
        self.path = path
        self.handler = Lazy(handler)
        self.name = name
        self.params = params or {}
        self.callbacks = callbacks or {}
        self.pattern = self.make_pattern(patterns)

    def match(self, path):
        if self.pattern:
            match = self.pattern.match(path)
            if match:
                params = {
                    param: self.callbacks.get(param, lambda x: x)(value)
                    for param, value in match.groupdict().items()
                }
                return path[match.end(0):], dict(self.params, **params)
        elif path.startswith(self.path):
            return path[len(self.path):], self.params

    def make_pattern(self, patterns=None):
        unescaped_path = self.path.replace('{{', '').replace('}}', '')
        placeholders = self.placeholder_re.findall(unescaped_path)
        if not placeholders:
            return
        patterns = patterns or {}
        pattern = re.escape(self.path.replace('{{', '{').replace('}}', '}'))
        for placeholder in placeholders:
            pattern = pattern.replace(
                r'\{{{placeholder}\}}'.format(placeholder=placeholder),
                r'(?P<{name}>{pattern})'.format(
                    name=placeholder,
                    pattern=patterns.get(placeholder, r'\w+')
                ),
            )
        return re.compile(pattern)

    def restore_path(self, **params):
        return self.path.format(**params)


class Routes(list):

    __slots__ = ('path', )

    def __init__(self, seq=(), path=''):
        super(Routes, self).__init__(seq)
        self.path = path

    def route(self, path, name=None, params=None, callbacks=None,
              patterns=None):
        def _decorator(handler):
            route = Route(self.path + path, handler,
                          name=name,
                          params=params,
                          callbacks=callbacks,
                          patterns=patterns)
            self.append(route)
            return handler
        return _decorator


class Header(collections.Mapping):

    __slots__ = 'value', 'params'

    def __init__(self, _value, **params):
        self.value = _value
        self.params = params

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return self.value == other

    def __ne__(self, other):
        return self.value != other

    def __str__(self):
        return self.stringify()

    def __bytes__(self):
        value = self.stringify()
        if isinstance(value, bytes):  # python 2.x
            return value
        return value.encode(encoding='latin1')

    def __getitem__(self, item):
        return self.params[item]

    def __iter__(self):
        return iter(self.params)

    def __len__(self):
        len(self.params)

    def stringify(self):
        if not self.params:
            return str(self.value)
        return '{value}; {params}'.format(
            value=self.value,
            params='; '.join(
                '%s=%s' % (attr_name, attr_value)
                for attr_name, attr_value in self.params.items()
            ),
        )
